from pyrb.model.account import Account
from pyrb.repository.account import AccountRepository


class AccountService:
    def __init__(self, account_repo: AccountRepository) -> None:
        self._account_repo = account_repo

    def set(self, account: Account) -> None:
        self._account_repo.set(account)

    def get(self) -> Account:
        return self._account_repo.get()



from typing import Any

import requests
from requests import Response

from pyrb.exceptions import APIClientError
from pyrb.model.account import EbestAccount
from pyrb.repository.brokerage.base.client import BrokerageAPIClient


class EbestAPIClient(BrokerageAPIClient):
    BASE_URL = "https://openapi.ebestsec.co.kr:8080"

    def __init__(self, account: EbestAccount) -> None:
        self._account = account

        self._access_token = self._issue_access_token()

    def send_request(self, method: str, path: str, **kwargs: Any) -> Response:
        URL = f"{self.BASE_URL}/{path}"
        headers = kwargs.get("headers", {})
        headers["authorization"] = f"Bearer {self._access_token}"
        kwargs["headers"] = headers

        response = requests.request(method, URL, **kwargs)

        # If token expired, renew and retry once
        if response.status_code == 401:  # Assuming 401 status code indicates an expired token
            self._access_token = self._issue_access_token()
            headers["authorization"] = f"Bearer {self._access_token}"
            response = requests.request(method, URL, **kwargs)

        self._raise_for_status(response)

        return response

    def _issue_access_token(self) -> str:
        path = "oauth2/token"
        url = f"{self.BASE_URL}/{path}"

        app_key = self._account.app_key
        app_secret = self._account.app_secret

        headers = {"content-type": "application/x-www-form-urlencoded"}
        params = {
            "grant_type": "client_credentials",
            "appkey": app_key,
            "appsecretkey": app_secret,
            "scope": "oob",
        }

        response = requests.post(url, headers=headers, params=params)

        self._raise_for_status(response)

        return response.json()["access_token"]

    def _raise_for_status(self, response: Response) -> None:
        try:
            response.raise_for_status()
        except requests.HTTPError:
            error_code = response.json()["rsp_cd"]
            error_msg = response.json()["rsp_msg"]
            status_code = response.status_code
            raise APIClientError(error_code, error_msg, status_code)



from pydantic import BaseModel

from pyrb.enums import OrderSide, OrderStatus, OrderType


class Order(BaseModel):
    symbol: str  # 종목코드
    price: int  # 주문가격
    quantity: int  # 주문수량
    side: OrderSide  # 매매구분
    order_type: OrderType  # 주문유형
    status: OrderStatus = OrderStatus.PENDING  # 주문상태

    def __str__(self) -> str:
        return f"{self.symbol}: {self.side} {self.quantity} shares @ {self.price}"



from math import floor

from pyrb.enums import (
    OrderSide,
    OrderStatus,
    OrderType,
)
from pyrb.exceptions import InsufficientFundsException, OrderPlacementError
from pyrb.model.order import Order
from pyrb.model.price import CurrentPrice
from pyrb.repository.brokerage.base.portfolio import Portfolio
from pyrb.repository.brokerage.context import RebalanceContext
from pyrb.service.strategy.base import Strategy


class Rebalancer:
    def __init__(self, context: RebalanceContext, strategy: Strategy) -> None:
        self._context = context
        self._strategy = strategy

    def prepare_orders(self, investment_amount: float) -> list[Order]:
        """
        Prepare a list of orders to rebalance the portfolio based on the given investment amount.
        If the investment amount is greater than the total value of the portfolio, an exception
        will be raised.

        Args:
            investment_amount : The amount of money to invest in the portfolio.

        Returns:
            list[Order]: A list of orders to rebalance the portfolio.
        """

        def _calculate_shares_to_trade(difference_in_amount: float, current_price: float) -> int:
            """Determine the number of shares to trade based on the difference in amount."""
            return floor(difference_in_amount / current_price)

        def _get_current_stock_amount(portfolio: Portfolio, stock: str) -> float:
            """Retrieve the current amount of a stock."""
            current_position = portfolio.get_position(stock)
            return current_position.total_amount if current_position else 0

        def _fetch_current_prices(
            context: RebalanceContext, weight_by_stock: dict[str, float]
        ) -> dict[str, CurrentPrice]:
            """Fetch the current prices of stocks in the portfolio and target stocks."""
            holding_symbols = context.portfolio.holding_symbols
            target_symbols = weight_by_stock.keys()
            whole_symbols = list(set(holding_symbols).union(target_symbols))
            current_prices = context.price_fetcher.get_current_prices(whole_symbols)
            return current_prices

        self._validate_investment_amount(investment_amount)

        weight_by_stock = self._strategy.create_target_weights()

        current_prices = _fetch_current_prices(self._context, weight_by_stock)
        orders: list[Order] = []

        for stock, weight in weight_by_stock.items():
            current_price = current_prices[stock].price
            current_amount = _get_current_stock_amount(self._context.portfolio, stock)

            target_amount = investment_amount * weight
            difference_in_amount = target_amount - current_amount
            shares_to_trade = _calculate_shares_to_trade(difference_in_amount, current_price)

            if shares_to_trade != 0:
                order_action = OrderSide.BUY if shares_to_trade > 0 else OrderSide.SELL
                orders.append(
                    Order(
                        symbol=stock,
                        price=current_price,
                        quantity=abs(shares_to_trade),
                        side=order_action,
                        order_type=OrderType.MARKET,
                    )
                )

        # 매도주문을 우선 제출
        orders.sort(key=lambda order: order.side == OrderSide.SELL, reverse=True)
        return orders

    def place_orders(self, orders: list[Order]) -> None:
        """
        Place a list of orders in the market.
        The status of each order will be updated based on the result of the order placement.

        Args:
            orders (list[Order]): A list of orders to be placed in the market.

        Returns:
            None
        """
        for order in orders:
            try:
                self._context.order_manager.place_order(order)
                order.status = OrderStatus.PLACED

            except OrderPlacementError:
                order.status = OrderStatus.REJECTED  # TODO: handle rejected orders

    def _validate_investment_amount(self, investment_amount: float) -> None:
        if investment_amount > self._context.portfolio.total_value:
            raise InsufficientFundsException(
                "Insufficient funds. The amount of your total asset is"
                f" {self._context.portfolio.total_value}"
            )



from abc import ABC, abstractmethod
from pathlib import Path

import toml

from pyrb.exceptions import InitializationError
from pyrb.model.account import Account, AccountFactory


class AccountRepository(ABC):
    @abstractmethod
    def set(self, account: Account) -> None: ...

    @abstractmethod
    def get(self) -> Account: ...


class LocalConfigAccountRepository(AccountRepository):
    def __init__(self, config_path: Path) -> None:
        self._config_path = config_path
        # create config file within a directory if not exists
        self._config_path.parent.mkdir(parents=True, exist_ok=True)
        self._config_path.touch(exist_ok=True)

    def set(self, account: Account) -> None:
        with open(self._config_path, "w") as f:
            f.write(account.to_toml())

    def get(self) -> Account:
        with open(self._config_path) as f:
            account_config = toml.loads(f.read())
            if not account_config:
                raise InitializationError("account is not set. Please set account first")

            brokerage = account_config.pop("brokerage")
            account = AccountFactory.create(brokerage=brokerage, **account_config)

            return account



from pyrb.model.account import Account, EbestAccount
from pyrb.repository.brokerage.base.client import BrokerageAPIClient
from pyrb.repository.brokerage.base.fetcher import PriceFetcher
from pyrb.repository.brokerage.base.order_manager import OrderManager
from pyrb.repository.brokerage.base.portfolio import Portfolio
from pyrb.repository.brokerage.ebest.client import EbestAPIClient
from pyrb.repository.brokerage.ebest.fetcher import EbestPriceFetcher
from pyrb.repository.brokerage.ebest.order_manager import EbestOrderManager
from pyrb.repository.brokerage.ebest.portfolio import EbestPortfolio


class BrokerageAPIClientFactory:
    def __init__(self) -> None: ...

    def create(self, account: Account) -> BrokerageAPIClient:
        match account:
            case EbestAccount():
                return EbestAPIClient(account)
            case _:
                raise NotImplementedError(f"Unsupported account: {account}")


class PortfolioFactory:
    def __init__(self) -> None: ...

    def create(self, brokerage_api_client: BrokerageAPIClient) -> Portfolio:
        match brokerage_api_client:
            case EbestAPIClient():
                return EbestPortfolio(brokerage_api_client)
            case _:
                raise NotImplementedError(f"Unsupported BrokerageAPIClient: {brokerage_api_client}")


class PriceFetcherFactory:
    def __init__(self) -> None: ...

    def create(self, brokerage_api_client: BrokerageAPIClient) -> PriceFetcher:
        match brokerage_api_client:
            case EbestAPIClient():
                return EbestPriceFetcher(brokerage_api_client)
            case _:
                raise NotImplementedError(f"Unsupported BrokerageAPIClient: {brokerage_api_client}")


class OrderManagerFactory:
    def __init__(self) -> None: ...

    def create(self, brokerage_api_client: BrokerageAPIClient) -> OrderManager:
        match brokerage_api_client:
            case EbestAPIClient():
                return EbestOrderManager(brokerage_api_client)
            case _:
                raise NotImplementedError(f"Unsupported BrokerageAPIClient: {brokerage_api_client}")



import abc

from pyrb.model.order import Order


class OrderManager(abc.ABC):
    def __init__(self) -> None: ...

    @abc.abstractmethod
    def place_order(self, order: Order) -> None:
        """
        Places an order with the brokerage. This method should be implemented by the
        concrete class.
        If the order fails to place, an OrderPlacementError should be raised.

        Args:
            order (Order): The order to place.

        Returns:
            None

        Raises:
            OrderPlacementError: If the order fails to place.
        """
        ...



from pathlib import Path
from typing import Annotated

import typer
from rich.console import Console
from rich.table import Table

from pyrb.controller.cli.account import APP_NAME
from pyrb.controller.cli.account import app as account_app
from pyrb.enums import AssetAllocationStrategyEnum, OrderSide, OrderStatus
from pyrb.model.order import Order
from pyrb.repository.account import LocalConfigAccountRepository
from pyrb.repository.brokerage.context import RebalanceContext, create_rebalance_context
from pyrb.service.account import AccountService
from pyrb.service.rebalance import Rebalancer
from pyrb.service.strategy.asset_allocate import (
    AssetAllocationStrtegyFactory,
)
from pyrb.service.strategy.explicit_target import (
    ExplicitTargetRebalanceStrategy,
    read_targets_from_source,
)
from pyrb.service.strategy.holding_portfolio import HoldingPortfolioRebalanceStrategy

app = typer.Typer()
app.add_typer(account_app, name="account")
console = Console()


@app.callback()
def callback() -> None:
    """Rebalance your portfolio"""
    ...


@app.command()
def holding_portfolio(
    investment_amount: Annotated[float, typer.Option(..., help="The total investment amount")],
) -> None:
    """
    Rebalances a holding portfolio with equal weights based on the specified options.
    """
    app_config_dir = Path(typer.get_app_dir(APP_NAME))
    accounts_config_path = app_config_dir / "accounts"

    account_service = AccountService(
        account_repo=LocalConfigAccountRepository(accounts_config_path)
    )

    account = account_service.get()
    context = create_rebalance_context(account)
    strategy = HoldingPortfolioRebalanceStrategy(context)
    rebalancer = Rebalancer(context, strategy)

    orders = rebalancer.prepare_orders(investment_amount)
    user_confirmation = _get_confirm_for_order_submit(context, orders)

    if user_confirmation:
        rebalancer.place_orders(orders)
    else:
        typer.echo("No orders were placed")

    _report_orders(orders)


@app.command()
def explicit_target(
    targets_source: Annotated[
        Path,
        typer.Option(
            ...,
            help=(
                "The source to read the target weights from. Supported file types: .csv, .json,"
                " .yaml"
            ),  # TODO: add docstring for each file structure.
            exists=True,
            file_okay=True,
            dir_okay=False,
            writable=False,
            readable=True,
            resolve_path=True,
        ),
    ],
    investment_amount: Annotated[float, typer.Option(..., help="The total investment amount")],
) -> None:
    """
    Rebalances a portfolio with explicit target weights from the specified source.
    Sum of target weights must be 1.0. If not, the weights will be normalized.

    """
    app_config_dir = Path(typer.get_app_dir(APP_NAME))
    accounts_config_path = app_config_dir / "accounts"

    account_service = AccountService(
        account_repo=LocalConfigAccountRepository(accounts_config_path)
    )

    account = account_service.get()
    context = create_rebalance_context(account)
    targets = read_targets_from_source(targets_source)
    strategy = ExplicitTargetRebalanceStrategy(targets)
    rebalancer = Rebalancer(context, strategy)

    orders = rebalancer.prepare_orders(investment_amount=investment_amount)
    user_confirmation = _get_confirm_for_order_submit(context, orders)

    if user_confirmation:
        rebalancer.place_orders(orders)
    else:
        typer.echo("No orders were placed")

    _report_orders(orders)


@app.command()
def asset_allocate(
    strategy: Annotated[AssetAllocationStrategyEnum, typer.Option(..., help="The strategy to use")],
    investment_amount: Annotated[float, typer.Option(..., help="The total investment amount")],
) -> None:
    """
    Rebalances a portfolio with the specified asset allocation strategy.

    """
    app_config_dir = Path(typer.get_app_dir(APP_NAME))
    accounts_config_path = app_config_dir / "accounts"

    account_service = AccountService(
        account_repo=LocalConfigAccountRepository(accounts_config_path)
    )

    account = account_service.get()
    context = create_rebalance_context(account)

    strategy = AssetAllocationStrtegyFactory().create(strategy)
    rebalancer = Rebalancer(context, strategy)

    orders = rebalancer.prepare_orders(investment_amount=investment_amount)
    user_confirmation = _get_confirm_for_order_submit(context, orders)

    if user_confirmation:
        rebalancer.place_orders(orders)
    else:
        typer.echo("No orders were placed")

    _report_orders(orders)


@app.command()
def portfolio() -> None:
    app_config_dir = Path(typer.get_app_dir(APP_NAME))
    accounts_config_path = app_config_dir / "accounts"

    account_service = AccountService(
        account_repo=LocalConfigAccountRepository(accounts_config_path)
    )

    account = account_service.get()
    context = create_rebalance_context(account)
    context.portfolio.refresh()

    table = Table(
        "Symbol",
        "Quantity",
        "Sellable Quantity",
        "Average Buy Price",
        "Total Amount",
        "Return (%)",
    )

    for position in context.portfolio.positions:
        table.add_row(
            position.symbol,
            str(position.quantity),
            str(position.sellable_quantity),
            str(position.average_buy_price),
            str(position.total_amount),
            str(position.rtn),
        )

    console.print(table)


def _get_confirm_for_order_submit(context: RebalanceContext, orders: list[Order]) -> bool:
    """Confirm orders to the user and return the user's confirmation."""
    table = Table(
        "Symbol",
        "Side",
        "Quantity",
        "Price",
        "Total Amount",
        "Current position value",
        "Expected position value",
    )

    for order in orders:
        total_amount = order.quantity * order.price
        current_position_value = context.portfolio.get_position_amount(order.symbol)
        expected_position_value = (
            current_position_value + total_amount
            if order.side == OrderSide.BUY
            else current_position_value - total_amount
        )

        table.add_row(
            order.symbol,
            order.side,
            str(order.quantity),
            str(order.price),
            str(total_amount),
            str(current_position_value),
            str(expected_position_value),
        )

    console.print(table)

    return typer.confirm("Do you want to place these orders?")


def _report_orders(orders: list[Order]) -> None:
    """Provides a summary of successful and failed orders."""
    for order in orders:
        if order.status == OrderStatus.PLACED:
            typer.echo(f"Successfully placed order: {order}")
        elif order.status == OrderStatus.REJECTED:
            typer.echo(f"Failed to place order: {order}")


if __name__ == "__main__":
    app()



import abc
from typing import Any

from requests import Response


class BrokerageAPIClient(abc.ABC):
    @abc.abstractmethod
    def send_request(self, method: str, path: str, **kwargs: Any) -> Response: ...



from abc import abstractmethod

from pyrb.enums import AssetAllocationStrategyEnum
from pyrb.service.strategy.base import Strategy


class AssetAllocationStrategy(Strategy):
    @abstractmethod
    def create_target_weights(self) -> dict[str, float]: ...


class AllWeatherKRStrategy(AssetAllocationStrategy):
    def __init__(self) -> None: ...

    def create_target_weights(self) -> dict[str, float]:
        return {
            "379800": 0.175,  # (주식) KODEX 미국S&P500TR
            "361580": 0.175,  # (주식) KBSTAR 200TR
            "411060": 0.15,  # (금) ACE KRX금현물
            "365780": 0.175,  # (국채) ACE 국고채10년
            "308620": 0.175,  # (국채) KODEX 미국채10년선물
            "272580": 0.15,  # (현금성 자산) TIGER 단기채권액티브
        }


class AssetAllocationStrtegyFactory:
    def create(self, strategy_type: AssetAllocationStrategyEnum) -> AssetAllocationStrategy:
        if strategy_type == AssetAllocationStrategyEnum.ALL_WEATHER_KR:
            return AllWeatherKRStrategy()
        else:
            raise ValueError(f"Unsupported strategy type: {strategy_type}")



from pydantic import BaseModel, PositiveFloat, PositiveInt


class Position(BaseModel):
    symbol: str  # 종목코드
    quantity: PositiveInt  # 보유수량
    sellable_quantity: PositiveInt  # 매도가능수량
    average_buy_price: PositiveFloat  # 매입단가
    total_amount: PositiveFloat  # 평가금액
    rtn: float  # 수익률



from pyrb.repository.brokerage.context import RebalanceContext
from pyrb.service.strategy.base import Strategy


class HoldingPortfolioRebalanceStrategy(Strategy):
    """
    A rebalancing strategy
    that creates target weights for a portfolio based on the current holdings.

    Args:
        portfolio (Portfolio): The portfolio to rebalance.
    """

    def __init__(self, context: RebalanceContext) -> None:
        self._context = context

    def create_target_weights(self) -> dict[str, float]:
        holding_symbols = self._context._portfolio.holding_symbols
        return {stock: 1 / len(holding_symbols) for stock in holding_symbols}



from typing import Any

from pydantic import NonNegativeFloat

from pyrb.model.position import Position
from pyrb.repository.brokerage.base.portfolio import Portfolio
from pyrb.repository.brokerage.ebest.client import EbestAPIClient


class EbestPortfolio(Portfolio):
    def __init__(self, api_client: EbestAPIClient) -> None:
        self._api_client = api_client
        self._serialized_portfolio: dict[str, Any] = self._fetch_portfolio()

    @property
    def total_value(self) -> NonNegativeFloat:
        return self._serialized_portfolio["t0424OutBlock"]["sunamt"]

    @property
    def positions(self) -> list[Position]:
        positions = [
            Position(
                symbol=item["expcode"],
                quantity=item["janqty"],
                sellable_quantity=item["mdposqt"],
                average_buy_price=item["pamt"],
                total_amount=item["appamt"],
                rtn=item["sunikrt"],
            )
            for item in self._serialized_portfolio["t0424OutBlock1"]
        ]

        return positions

    @property
    def holding_symbols(self) -> list[str]:
        return [position.symbol for position in self.positions]

    def get_position(self, symbol: str) -> Position | None:
        return next((position for position in self.positions if position.symbol == symbol), None)

    def get_position_amount(self, symbol: str) -> NonNegativeFloat:
        position = self.get_position(symbol)
        return position.total_amount if position else 0

    def refresh(self) -> None:
        self._serialized_portfolio = self._fetch_portfolio()

    def _fetch_portfolio(self) -> dict[str, Any]:
        """주식잔고2 TR을 조회합니다.

        see: https://openapi.ebestsec.co.kr/apiservice?group_id=73142d9f-1983-48d2-8543-89b75535d34c&api_id=37d22d4d-83cd-40a4-a375-81b010a4a627
        """
        path = "stock/accno"
        content_type = "application/json; charset=UTF-8"

        headers = {"content-type": content_type, "tr_cd": "t0424", "tr_cont": "N"}
        body = {
            "t0424InBlock": {
                "prcgb": "",
                "chegb": "",
                "dangb": "",
                "charge": "",
                "cts_expcode": "",
            }
        }

        response = self._api_client.send_request("POST", path, headers=headers, json=body)
        return response.json()



from pyrb.model.price import CurrentPrice
from pyrb.repository.brokerage.base.fetcher import PriceFetcher
from pyrb.repository.brokerage.ebest.client import EbestAPIClient


class EbestPriceFetcher(PriceFetcher):
    def __init__(self, api_client: EbestAPIClient) -> None:
        self._api_client = api_client

    def get_current_price(self, symbol: str) -> CurrentPrice:
        return self.get_current_prices([symbol])[symbol]

    def get_current_prices(self, symbols: list[str]) -> dict[str, CurrentPrice]:
        path = "stock/market-data"
        content_type = "application/json; charset=UTF-8"

        headers = {"content-type": content_type, "tr_cd": "t8407", "tr_cont": "N"}
        body = {
            "t8407InBlock": {
                "nrec": len(symbols),  # 조회할 종목 수
                "shcode": "".join(symbols),  # 종목코드
            }
        }

        response = self._api_client.send_request("POST", path, headers=headers, json=body)

        res = response.json()
        current_prices = {
            item["shcode"]: CurrentPrice(symbol=item["shcode"], price=item["price"])
            for item in res["t8407OutBlock1"]
        }
        return current_prices



import csv
import json
from abc import ABC, abstractmethod
from pathlib import Path

import yaml

from pyrb.exceptions import InvalidTargetError
from pyrb.service.strategy.base import Strategy


class ExplicitTargetRebalanceStrategy(Strategy):
    """
    A rebalancing strategy that uses explicit target weights for each asset.

    Args:
        targets (dict[str, float]): A dictionary mapping asset symbols to target weights.

    """

    def __init__(self, targets: dict[str, float]) -> None:
        self._targets = targets

    def create_target_weights(self) -> dict[str, float]:
        return self._targets


def read_targets_from_source(source: Path) -> dict[str, float]:
    """
    Reads the target weights from the specified source.
    currently surpports sources are:
    - csv file
    - json file
    - yaml file

    Args:
        source (Path): The source to read the target weights from.

    Returns:
        dict[str, float]: A dictionary mapping asset symbols to target weights.

    """
    match source.suffix:
        case ".csv":
            return CSVTargetReader().read(source)
        case ".json":
            return JSONTargetReader().read(source)
        case ".yaml":
            return YAMLTargetReader().read(source)
        case _:
            raise InvalidTargetError(f"Unsupported source file type: {source.suffix}")


class TargetReader(ABC):
    """
    Abstract class representing a reader for target allocations.
    Provides functionality to read, validate, and normalize target allocations.

    """

    def read(self, source: Path) -> dict[str, float]:
        """
        Reads, validates, and normalizes target allocations from the given source.

        Args:
            source (Path): The source to read the target allocations from.

        Returns:
            dict[str, float]: A dictionary mapping asset symbols to target weights.

        Raises:
            InvalidTargetError: If the target allocations are invalid.
        """
        targets = self._read_targets(source)
        self._validate_targets(targets)
        normalized_targets = self._normalize_targets(targets)
        return normalized_targets

    def _validate_targets(self, targets: dict[str, float]) -> None:
        def _validate_symbol(symbol: str) -> None:
            if not isinstance(symbol, str):
                raise InvalidTargetError("Target symbol must be a string")

        def _validate_weight(weight: float) -> None:
            if not isinstance(weight, float):
                raise InvalidTargetError("Target weight must be a float")

            if weight <= 0:
                raise InvalidTargetError("Target weight must be positive")

        if not isinstance(targets, dict):
            raise InvalidTargetError("Targets must be a dictionary")

        for symbol, weight in targets.items():
            _validate_symbol(symbol)
            _validate_weight(weight)

    def _normalize_targets(self, targets: dict[str, float]) -> dict[str, float]:
        total_weight = sum(targets.values())
        normalized_targets = {symbol: weight / total_weight for symbol, weight in targets.items()}
        return normalized_targets

    @abstractmethod
    def _read_targets(self, source: Path) -> dict[str, float]:
        """
        Read the targets from the given source.

        Args:
            source (Path): The source to read the target weights from.

        Returns:
            dict[str, float]: A dictionary mapping asset symbols to target weights.
        """
        ...


class CSVTargetReader(TargetReader):
    def _read_targets(self, source: Path) -> dict[str, float]:
        with open(source, newline="") as csvfile:
            reader = csv.DictReader(csvfile)
            targets = {row["symbol"]: float(row["weight"]) for row in reader}
            return targets


class JSONTargetReader(TargetReader):
    def _read_targets(self, source: Path) -> dict[str, float]:
        with open(source) as jsonfile:
            targets = json.load(jsonfile)
            return targets


class YAMLTargetReader(TargetReader):
    def _read_targets(self, source: Path) -> dict[str, float]:
        with open(source) as yamlfile:
            targets = yaml.safe_load(yamlfile)
            return targets



from abc import ABC, abstractmethod


class Strategy(ABC):
    @abstractmethod
    def create_target_weights(self) -> dict[str, float]:
        """
        Returns a dictionary with asset symbols as keys and target allocation percentages as values.
        """
        ...



from requests import HTTPError

from pyrb.enums import OrderType
from pyrb.exceptions import OrderPlacementError
from pyrb.model.order import Order
from pyrb.repository.brokerage.base.order_manager import OrderManager
from pyrb.repository.brokerage.ebest.client import EbestAPIClient


class EbestOrderManager(OrderManager):
    _order_type_mapping: dict[OrderType, str] = {
        OrderType.LIMIT: "00",
        OrderType.MARKET: "03",
        OrderType.CONDITIONAL_LIMIT: "05",
        OrderType.BEST_LIMIT: "06",
        OrderType.IMMEDIATE_LIMIT: "07",
        OrderType.PREOPENING_SESSION_LAST: "61",
        OrderType.AFTER_HOURS_LAST: "81",
        OrderType.AFTER_HOURS_SINGLE: "82",
    }

    def __init__(self, api_client: EbestAPIClient) -> None:
        self._api_client = api_client

    def place_order(self, order: Order) -> None:
        path = "stock/order"
        content_type = "application/json; charset=UTF-8"

        headers = {"content-type": content_type, "tr_cd": "CSPAT00601", "tr_cont": "N"}
        body = {
            "CSPAT00601InBlock1": {
                "IsuNo": order.symbol,
                "OrdQty": order.quantity,
                "OrdPrc": order.price,
                "BnsTpCode": "2" if order.side == "BUY" else "1",
                "OrdprcPtnCode": self._order_type_mapping[order.order_type],
                "MgntrnCode": "000",
                "LoanDt": "",
                "OrdCndiTpCode": "0",
            }
        }

        try:
            self._api_client.send_request("POST", path, headers=headers, json=body)
        except HTTPError as e:
            raise OrderPlacementError(e)



from abc import ABC
from typing import Annotated, Any

import toml
from pydantic import BaseModel, Field

from pyrb.enums import BrokerageType


class Account(BaseModel, ABC):
    brokerage: Annotated[BrokerageType, Field(...)]

    def to_toml(self) -> str:
        model_dict = self.model_dump()
        model_dict.update({"brokerage": self.brokerage.value})
        return toml.dumps(model_dict)


class EbestAccount(Account):
    app_key: Annotated[str, Field(...)]
    app_secret: Annotated[str, Field(...)]


class AccountFactory:
    @staticmethod
    def create(brokerage: BrokerageType, **kwargs: Any) -> Account:
        if brokerage == BrokerageType.EBEST.value:
            return EbestAccount(brokerage=brokerage, **kwargs)
        raise ValueError(f"brokerage {brokerage} is not supported")



class PyRbException(Exception): ...


class InsufficientFundsException(PyRbException): ...


class PaperTradingSettingError(PyRbException): ...


class OrderPlacementError(PyRbException): ...


class InvalidTargetError(PyRbException): ...


class InitializationError(PyRbException): ...


class APIClientError(PyRbException):
    def __init__(self, client_error_code: str, client_error_message: str, status_code: int) -> None:
        self.client_error_code = client_error_code
        self.client_error_message = client_error_message
        self.status_code = status_code



import abc

from pydantic import NonNegativeFloat

from pyrb.model.position import Position


class Portfolio(abc.ABC):
    def __init__(self) -> None: ...

    @property
    @abc.abstractmethod
    def total_value(self) -> NonNegativeFloat:
        """Calculates the total value of the portfolio.
        total value = cash + sum of the market value of all positions

        Returns:
            NonNegativeFloat: The total value of the portfolio.
        """
        ...

    @property
    @abc.abstractmethod
    def positions(self) -> list[Position]:
        """Returns a list of all positions in the portfolio.

        Returns:
            list[Position]: A list of all positions in the portfolio.
        """
        ...

    @property
    @abc.abstractmethod
    def holding_symbols(self) -> list[str]:
        """Returns a list of symbols for all positions in the portfolio.

        Returns:
            list[str]: A list of symbols for all positions in the portfolio.
        """
        ...

    @abc.abstractmethod
    def get_position(self, symbol: str) -> Position | None:
        """Returns the position object for the given symbol.

        Args:
            symbol : The symbol of the position to retrieve.

        Returns:
            Position | None: The position object for the given symbol,
                             if the symbol is not found, None will be returned.
        """
        ...

    @abc.abstractmethod
    def get_position_amount(self, symbol: str) -> NonNegativeFloat:
        """Returns the total amount of the position for the given symbol.
        If the symbol is not found, 0 will be returned.

        Args:
            symbol : The symbol of the position to retrieve.

        Returns:
            NonNegativeFloat: The total amount of the position for the given symbol.
        """
        ...

    @abc.abstractmethod
    def refresh(self) -> None:
        """Refreshes the portfolio object."""
        ...



from enum import StrEnum


class BrokerageType(StrEnum):
    EBEST = "ebest"


class OrderType(StrEnum):
    LIMIT = "LIMIT"  # 지정가
    MARKET = "MARKET"  # 시장가
    CONDITIONAL_LIMIT = "CONDITIONAL_LIMIT"  # 조건부지정가
    BEST_LIMIT = "BEST_LIMIT"  # 최유리지정가
    IMMEDIATE_LIMIT = "IMMEDIATE_LIMIT"  # 최우선지정가
    PREOPENING_SESSION_LAST = "PREOPENING_SESSION_LAST"  # 장개시전시간외종가
    AFTER_HOURS_LAST = "AFTER_HOURS_LAST"  # 시간외종가
    AFTER_HOURS_SINGLE = "AFTER_HOURS_SINGLE"  # 시간외단일가


class OrderSide(StrEnum):
    BUY = "BUY"
    SELL = "SELL"


class OrderStatus(StrEnum):
    PENDING = "PENDING"
    PLACED = "PLACED"
    REJECTED = "REJECTED"
    CANCELLED = "CANCELLED"


class AssetAllocationStrategyEnum(StrEnum):
    ALL_WEATHER_KR = "all-weather-kr"



from pathlib import Path
from typing import Annotated

import typer

from pyrb.enums import BrokerageType
from pyrb.model.account import AccountFactory
from pyrb.repository.account import LocalConfigAccountRepository
from pyrb.service.account import AccountService

app = typer.Typer()
APP_NAME = "pyrb"  # TODO: parse from pyproject.toml and move to constants.py


@app.command("set")
def set(
    app_key: Annotated[str, typer.Option(..., help="app key for the brokerage", prompt=True)],
    app_secret: Annotated[str, typer.Option(..., help="app secret for the brokerage", prompt=True)],
    brokerage: Annotated[
        BrokerageType, typer.Option(help="brokerage type", case_sensitive=False)
    ] = BrokerageType.EBEST,
) -> None:
    app_config_dir = Path(typer.get_app_dir(APP_NAME))
    accounts_config_path = app_config_dir / "accounts"

    account_service = AccountService(
        account_repo=LocalConfigAccountRepository(accounts_config_path)
    )

    account = AccountFactory.create(brokerage, app_key=app_key, app_secret=app_secret)
    account_service.set(account=account)



from pyrb.model.account import Account
from pyrb.repository.brokerage.base.fetcher import PriceFetcher
from pyrb.repository.brokerage.base.order_manager import OrderManager
from pyrb.repository.brokerage.base.portfolio import Portfolio
from pyrb.repository.brokerage.factory import (
    BrokerageAPIClientFactory,
    OrderManagerFactory,
    PortfolioFactory,
    PriceFetcherFactory,
)


class RebalanceContext:
    def __init__(
        self, portfolio: Portfolio, price_fetcher: PriceFetcher, order_manager: OrderManager
    ) -> None:
        self._portfolio = portfolio
        self._price_fetcher = price_fetcher
        self._order_manager = order_manager

    @property
    def portfolio(self) -> Portfolio:
        return self._portfolio

    @property
    def price_fetcher(self) -> PriceFetcher:
        return self._price_fetcher

    @property
    def order_manager(self) -> OrderManager:
        return self._order_manager


def create_rebalance_context(account: Account) -> RebalanceContext:
    brokerage_api_client = BrokerageAPIClientFactory().create(account)

    portfolio = PortfolioFactory().create(brokerage_api_client)
    price_fetcher = PriceFetcherFactory().create(brokerage_api_client)
    order_manager = OrderManagerFactory().create(brokerage_api_client)

    rebalance_context = RebalanceContext(portfolio, price_fetcher, order_manager)
    return rebalance_context



import abc

from pyrb.model.price import CurrentPrice

# TODO: 우선매도잔량, 우선매수잔량 추가


class PriceFetcher(abc.ABC):
    def __init__(self) -> None: ...

    @abc.abstractmethod
    def get_current_price(self, symbol: str) -> CurrentPrice:
        """특정 종목의 현재가를 조회합니다.

        Args:
            symbol: 종목코드

        Returns:
            CurrentPrice 객체
        """
        ...

    @abc.abstractmethod
    def get_current_prices(self, symbols: list[str]) -> dict[str, CurrentPrice]:
        """복수 종목의 현재가를 조회합니다.

        Args:
            symbols: 종목코드 리스트

        Returns:
            키는 종목코드, 값은 CurrentPrice 객체로 매핑된 딕셔너리
        """
        ...
